# PyWCSTools - a simple SWIG wrapping of WCSTools (http://tdc-www.harvard.edu/software/wcstools/) by Doug Mink
#
# (c) 2007, 2008 Matt Hilton
#
# See the README file for information on usage and distribution.