"""
astLib - python astronomy modules

(c) 2007-2012 Matt Hilton

(c) 2013-2025 Matt Hilton & Steven Boada

U{https://astlib.readthedocs.io}

See the README file for information on usage and installation.
See the LICENSE file for information on distribution.

"""

__all__=['astCalc', 'astCoords', 'astImages', 'astPlots', 'astStats', 'astWCS', 'astSED']
__version__ = '0.13.2'
